"""This module provides the base class for drugref API database adapters
All drugref adapters should inherit from this base class
(c)2005 by Dr Horst Herb <hherb@drugref.org>
This is free software released under the GPL license version 2 or newer
Details of the GPL license condtion can be obtained at http://www.gnu.org/licenses/gpl.txt
"""

class NotImplementedError:
	"""This error is thrown whenever somebody tries to access a API method that has not been
	implemented for that particular database"""
	pass

class drugrefapi:
	"""Base class for all drugref API backend adapters"""
	def __init__(self, host='localhost', port=8123, database='drugref', user='drugref', pwd='drugref', backend='postgres'):
		"""The vast majority of adapters will probably use a SQL server backend. Adapters that don't need such
		connection parameters, can simply ignore them."""
		self._name = "generic adapter"
		self._version = '0.2'
		self._versiondate = '11.01.2005'
		self._params={}
		self._params['host']=host
		self._params['port']=port
		self._params['database']=database
		self._params['user']=user
		self._params['pwd']=pwd
		self._params['backend']=backend
		self._dbapi = None
		self._db = None
		self._wildcard = "*"
		self._languages = ['?']
		self._countries = ['?']
		if backend == 'postgres':
			self.wildcard = "%"
			self._notcasesensitive = 'ilike'
		self._queries = {}	#available queries; key is the query name, value is a tuple
					#containing  the actual query string, and the required search expression
					#variables as list of key:value pairs with keys being the variable names,
		self._cache = {}	#cached query results.
					#key is the provided data item, values is a tuple containing it's value
					#and the primary key used to retrieve it
		self._provided = {}	#data items that are provided by this drugref adapter
					#key is data item, value is the query name that can retrieve that item
		self._required = {}	#search expressions and primary keys that produced cached query results;
					#key is the query name, value is the search expression or primary key
		self._funcs = {}

	def addquery(self, query, requires, provides):
		"""register a query that will allow to find and retrieve attributes.
		query: the query string, with parameter placeholdrs as per Python DBAPI 2.0
		requires: symbolic search expression keyword, as described in the drugref user manual
		provides: list of attribute names as described in the drugref user manual
		RETURNS: nothing"""
		for provided in provides:
			print "registering attribute", provided, "] for ", self._name
			self._provided[provided]={'query' : query, 'requires' : requires}

	def addfunc(self,func, requires, provides):
		"""register a function that will allow to find and retrieve attributes.
		func: a function that accepts a "key" as parameter and returns a list of lists of key:value pairs
		requires: symbolic search expression keyword, as described in the drugref user manual
		provides: list of attribute names as described in the drugref user manual
		RETURNS: nothing"""
		for provided in provides:
			print "registering attribute [", provided, "] for ", self._name
			self._provided[provided]={'func' : func, 'requires' : requires}


	def get(self, attribute, key):
		"""Access a drugref attribute as identified by parameter "key"
		Possible attributes and keys have to be looked up in the drugref
		user manual.
		RETURNS: list of lists of key:value pairs
		"""
                print "HERE IN GET"
		try:
			#check whether we have this attribute in cache
			attr = self._cache[attribute]
			#check whether the cached attribute is assciated with the identifying key
			if attr[1]== key:
				return attr[0]
		except:
			pass
		try:
			#check whether we can provide this attribute in principle (= service provider available)
			queryinfo = self._provided[attribute]
		except:
			return {'Error':'Attribute "%s" not available' % attribute}
		try:
			#if service provider available, try fiurst to retrieve the attribute by registered function
			results = queryinfo['func'](key)
                #else check whether we can retrieve the attribute via registered query
		except KeyError:
			try:
				self._connect()
				query = queryinfo['query']
				params = {queryinfo['requires'] : key}
				results = self._runquery(query, params)
			except:
				results=None
		if not results:
			return {'Error' : 'Not Found'}
		if queryinfo['requires'].startswith('_search'):
			#we are dealing with a list of result sets, not identified by a single primary key
			#thus no point ion caching anythying
			return results
		else:
			#let's fill the cache with the attributes
			#it should have been a primary key search, so there should only be one result
			assert(len(results)==1)
			result = results[0]
			for attribute in result:
				try:
					self._cache[attribute] = (result[attribute], key) 
				except:
					#this attribute is not meant to be exposed, so just skip it
					pass
		try:
			attr = self._cache[attribute]
			if attr[1]== key:
				return attr[0]
		except:
			pass
		return {'Error' : 'Not Found'}
			
		
	
	def _connect(self):
		"""connect to the database backend.
		RETURNS: nothing"""
		if self._params['backend'] == 'postgres':
			if not self._dbapi:
				import pyPgSQL.PgSQL as dbapi
				self._dbapi = dbapi
			if not self._db:
				self._db = self._dbapi.connect(database = self._params['database'])
			
	def _runquery(self, query, params=None):
		""" run the query "query" on a dbapi 2.0 compliant backend adapter,
		optionally using the query parameters "params".
		RETURNS: list of result sets"""
		self._connect()
		cur = self._db.cursor()
		if params:
			cur.execute(query, params)
		else:
			cur.execute(query)
		return cur.fetchall()
			
	def get_wildcard(self):
		"""get the wildcard that can be used in queries to represent any number of any characters
		RETURNS: character"""
		return self._wildcard

	def get_name(self):
		"""identify ourselves
		RETURNS: string"""
		return self._name

	def get_version(self):
		return self._version
				
	def list_capabilities(self):
		"""list all attributes we can provide
		RETURNS: list of strings"""
		return self._provided
		
	
if __name__ == "__main__":
	def testfunc(key):
		if key == 'thekey':
			return [{'attribute':'value', 'attribute2':'value2'}]
			
	def testsearch(key):
		return [{'found': 'one', 'key' : 1}, {'found': 'two', 'key' : 2}]
			
	a = drugrefapi()
	a.addfunc(testfunc, 'key', ['attribute', 'attribute2'])
	a.addfunc(testsearch, '_search_key', ['found', 'key'])
	print a.get('attribute', 'thekey')
	print a.get('attribute2', 'thekey')
	print a.get('found', 'test')
